// si_Functions.cpp - space invaders opengl

#define _CRT_SECURE_NO_WARNINGS			// For 4996 warning
#include "si_Main.h"

//////////////////////////////////////////////////////////////////////////////////////
//
// FONCTIONS => Ge une collision entre 2 rectangles (2 sprites)
//
//
int collision (float pg, float pd, float ph, float pb, float sg, float sd, float sh, float sb)
{
	if (
		( ((pg < sg) && (sg < pd)) || ((pg < sd) && (sd < pd)) )
		&&
		( ((pb < sb) && (sb < ph)) || ((pb < sh) && (sh < ph)) )
		)
		return 1;

	else if (
		( ((sg < pg) && (pg < sd)) || ((sg < pd) && (pd < sd)) )
		&&
		( ((sb < pb) && (pb < sh)) || ((sb < ph) && (ph < sh)) )
		)
		return 1;

	else return 0;
}

//////////////////////////////////////////////////////////////////////////////////////
//
// FONCTIONS => Ge le dai nessaire entre 2 frappes de Touch
//
//
int key_wait(int milliseconds, float & key_timer)		// ge les ritions de frappes de Touchs
{
	if ((TimerGetTime() - key_timer) < milliseconds)	// le temps oulentre 2 pressions
		return 0;										// d'une Touch doit re > milliseconds
	else
	{
		key_timer = TimerGetTime();
		return 1;
	}
}


//////////////////////////////////////////////////////////////////////////////////////
//
// FONCTIONS => Courbe gale de Bier. Retourne le point (donc les coordonns) situsur
//              la courbe form des n points de contres situ l'adresse p l'intant mu.
//
// FORMAT    => la structure d'un point:    typedef struct {float x;float y;float z;}point3f;
//              les points de contres:    dlar par: point3f control_points[n+1];
//                                          dinis par:  control_points[0].x=... jusqu'control_points[n].z=...
//              le point retourn          point3f return_point;
//              le temps:                   mu compris entre 0 et 1 (0 <= mu <= 1)
//
// APPEL     => return_point = Bezier(control_points, 5, t); (exemple pour 6 points de contre)
//
point3f Bezier(point3f *p,int n,double mu)
{
   int k,kn,nn,nkn;
   double blend,muk,munk;
   point3f b = {0.0,0.0,0.0};

   // cas particulier pour mu = 1
   if (mu == 1)
   {
	b.x = p[n].x;
	b.y = p[n].y;
	b.z = p[n].z;
	return(b);
   }

   // cas gal   
   muk = 1;
   munk = pow(1-mu,(double)n);

   for (k=0;k<=n;k++) {
      nn = n;
      kn = k;
      nkn = n - k;
      blend = muk * munk;
      muk *= mu;
      munk /= (1-mu);
      while (nn >= 1) {
         blend *= nn;
         nn--;
         if (kn > 1) {
            blend /= (double)kn;
            kn--;
         }
         if (nkn > 1) {
            blend /= (double)nkn;
            nkn--;
         }
      }
      b.x += p[k].x * blend;
      b.y += p[k].y * blend;
      b.z += p[k].z * blend;
   }

   return(b);
}

//////////////////////////////////////////////////////////////////////////////////////
//
// FONCTIONS => Retourne le hiscore
//
//
int get_hiscore()
{
	// LECTURE DU HISCORE DANS LA BASE DE REGISTRE
	HKEY key;
	DWORD disposition;

	// check key
	if (RegOpenKeyEx(HKEY_CURRENT_USER, "SOFTWARE\\SpaceInvaders-OpenGL\\", 0, KEY_READ, &key) == ERROR_SUCCESS)
	{
		// read key value
		int  nombre = 0;

		DWORD type = REG_DWORD;
		DWORD size = sizeof(DWORD);

		RegQueryValueEx(key, "hiscore", NULL, &type, (LPBYTE)&nombre, &size);
		return nombre;
	}
	else
	{
		// create key
		RegCreateKeyEx(HKEY_CURRENT_USER, "SOFTWARE\\SpaceInvaders-OpenGL\\", 0, "", REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS, NULL, &key, &disposition);
		return 0;
	}
	RegCloseKey(key);
}

//////////////////////////////////////////////////////////////////////////////////////
//
// FONCTIONS => Inscrit le hiscore
//
//
void set_hiscore(int score)
{
	// SAUVEGARDE DU HISCORE DANS LA BASE DE REGISTRE
	HKEY key;

	int current_hiscore = 0;
	current_hiscore = get_hiscore();

	if (score < current_hiscore)
		return;

	// check key
	if (RegOpenKeyEx(HKEY_CURRENT_USER, "SOFTWARE\\SpaceInvaders-OpenGL\\", 0, KEY_WRITE, &key) == ERROR_SUCCESS)
	{
		RegSetValueEx(key, "hiscore", 0, REG_DWORD, (LPBYTE) &score, sizeof(DWORD));
	}
	RegCloseKey(key);
}

//////////////////////////////////////////////////////////////////////////////////////
//
// FONCTIONS => La carte graphique est-elle acc ?
//
//
bool IsCardAccelerated()
{
   //structure that holds info about a pixel format
   PIXELFORMATDESCRIPTOR pfd;
   //HW accelerated yes or not
   bool isAccelerated = false;
  
   //Very important to set this variable.. Windows seems
   //not to be able to do this itself;)
   pfd.nSize = sizeof(PIXELFORMATDESCRIPTOR);
   
   //Device Context of the window you want to draw in
   //must be the same as the one you use to draw GL
   //HDC hDC = GetDC(g_window->hWnd);
   HDC hDC = GetDC(HWND_DESKTOP);
 
   //number of available pixel formats
   int nFormatCount = DescribePixelFormat(hDC, 1, 0, NULL);
 
   //Go through all available pixel formats and check..
   for(int i = 1; i <= nFormatCount; i++)
   {
      //Get description of pixel format
      DescribePixelFormat(hDC, i, pfd.nSize, &pfd);
 
      //Not generic?
      if (!(pfd.dwFlags & PFD_GENERIC_FORMAT))
      {
         //It's HW accelerated!
         isAccelerated = true;
         //We can stop here, as there is at least one acc. pixel format
         break;
      }
   }
  
   return isAccelerated;
}

//////////////////////////////////////////////////////////////////////////////////////
//
// FONCTIONS => Hyperliens (Copyright 2002 Neal Stublen, http://www.awesoftware.com)
//              Appel par : ConvertStaticToHyperlink(hwndDlg, IDC_MY_CONTROL); dans WM_INITDIALOG
//              et ShellExecute() dans case IDC_MY_CONTROL:
//
#define PROP_ORIGINAL_FONT		TEXT("_Hyperlink_Original_Font_")
#define PROP_ORIGINAL_PROC		TEXT("_Hyperlink_Original_Proc_")
#define PROP_STATIC_HYPERLINK	TEXT("_Hyperlink_From_Static_")
#define PROP_UNDERLINE_FONT		TEXT("_Hyperlink_Underline_Font_")

LRESULT CALLBACK _HyperlinkParentProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	WNDPROC pfnOrigProc = (WNDPROC) GetProp(hwnd, PROP_ORIGINAL_PROC);

	switch (message)
	{
	case WM_CTLCOLORSTATIC:
		{
			HDC hdc = (HDC) wParam;
			HWND hwndCtl = (HWND) lParam;

			BOOL fHyperlink = (NULL != GetProp(hwndCtl, PROP_STATIC_HYPERLINK));
			if (fHyperlink)
			{
				LRESULT lr = CallWindowProc(pfnOrigProc, hwnd, message, wParam, lParam);
				SetTextColor(hdc, RGB(0, 0, 0));
				return lr;
			}

			break;
		}
	case WM_DESTROY:
		{
			SetWindowLong(hwnd, GWL_WNDPROC, (LONG) pfnOrigProc);
			RemoveProp(hwnd, PROP_ORIGINAL_PROC);
			break;
		}
	}
	return CallWindowProc(pfnOrigProc, hwnd, message, wParam, lParam);
}

LRESULT CALLBACK _HyperlinkProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	WNDPROC pfnOrigProc = (WNDPROC) GetProp(hwnd, PROP_ORIGINAL_PROC);

	switch (message)
	{
	case WM_DESTROY:
		{
			SetWindowLong(hwnd, GWL_WNDPROC, (LONG) pfnOrigProc);
			RemoveProp(hwnd, PROP_ORIGINAL_PROC);

			HFONT hOrigFont = (HFONT) GetProp(hwnd, PROP_ORIGINAL_FONT);
			SendMessage(hwnd, WM_SETFONT, (WPARAM) hOrigFont, 0);
			RemoveProp(hwnd, PROP_ORIGINAL_FONT);

			HFONT hFont = (HFONT) GetProp(hwnd, PROP_UNDERLINE_FONT);
			DeleteObject(hFont);
			RemoveProp(hwnd, PROP_UNDERLINE_FONT);

			RemoveProp(hwnd, PROP_STATIC_HYPERLINK);

			break;
		}
	case WM_MOUSEMOVE:
		{
			if (GetCapture() != hwnd)
			{
				HFONT hFont = (HFONT) GetProp(hwnd, PROP_UNDERLINE_FONT);
				SendMessage(hwnd, WM_SETFONT, (WPARAM) hFont, FALSE);
				InvalidateRect(hwnd, NULL, FALSE);
				SetCapture(hwnd);
			}
			else
			{
				RECT rect;
				GetWindowRect(hwnd, &rect);

				POINT pt = { LOWORD(lParam), HIWORD(lParam) };
				ClientToScreen(hwnd, &pt);

				if (!PtInRect(&rect, pt))
				{
					HFONT hFont = (HFONT) GetProp(hwnd, PROP_ORIGINAL_FONT);
					SendMessage(hwnd, WM_SETFONT, (WPARAM) hFont, FALSE);
					InvalidateRect(hwnd, NULL, FALSE);
					ReleaseCapture();
				}
			}
			break;
		}
	case WM_SETCURSOR:
		{
			// Since IDC_HAND is not available on all operating systems,
			// we will load the arrow cursor if IDC_HAND is not present.
			HCURSOR hCursor = LoadCursor(NULL, MAKEINTRESOURCE(32649)); // Hand cursor
			if (NULL == hCursor)
			{
				hCursor = LoadCursor(NULL, MAKEINTRESOURCE(IDC_ARROW));
			}
			SetCursor(hCursor);
			return TRUE;
		}
	}

	return CallWindowProc(pfnOrigProc, hwnd, message, wParam, lParam);
}

BOOL ConvertStaticToHyperlink(HWND hwndCtl)
{
	// Subclass the parent so we can color the controls as we desire.

	HWND hwndParent = GetParent(hwndCtl);
	if (NULL != hwndParent)
	{
		WNDPROC pfnOrigProc = (WNDPROC) GetWindowLong(hwndParent, GWL_WNDPROC);
		if (pfnOrigProc != _HyperlinkParentProc)
		{
			SetProp(hwndParent, PROP_ORIGINAL_PROC, (HANDLE) pfnOrigProc);
			SetWindowLong(hwndParent, GWL_WNDPROC, (LONG) (WNDPROC) _HyperlinkParentProc);
		}
	}

	// Make sure the control will send notifications.

	DWORD dwStyle = GetWindowLong(hwndCtl, GWL_STYLE);
	SetWindowLong(hwndCtl, GWL_STYLE, dwStyle | SS_NOTIFY);

	// Subclass the existing control.

	WNDPROC pfnOrigProc = (WNDPROC) GetWindowLong(hwndCtl, GWL_WNDPROC);
	SetProp(hwndCtl, PROP_ORIGINAL_PROC, (HANDLE) pfnOrigProc);
	SetWindowLong(hwndCtl, GWL_WNDPROC, (LONG) (WNDPROC) _HyperlinkProc);

	// Create an updated font by adding an underline.

	HFONT hOrigFont = (HFONT) SendMessage(hwndCtl, WM_GETFONT, 0, 0);
	SetProp(hwndCtl, PROP_ORIGINAL_FONT, (HANDLE) hOrigFont);

	LOGFONT lf;
	GetObject(hOrigFont, sizeof(lf), &lf);
	lf.lfUnderline = TRUE;

	HFONT hFont = CreateFontIndirect(&lf);
	SetProp(hwndCtl, PROP_UNDERLINE_FONT, (HANDLE) hFont);

	// Set a flag on the control so we know what color it should be.

	SetProp(hwndCtl, PROP_STATIC_HYPERLINK, (HANDLE) 1);

	return TRUE;
}

BOOL ConvertStaticToHyperlink(HWND hwndParent, UINT uiCtlId)
{
	return ConvertStaticToHyperlink(GetDlgItem(hwndParent, uiCtlId));
}

void fontColor()
{
	if (game_3d)
	{
		//fontColor(0,0,0);
		//fontColor(0.5,0.5,0.5);
		//fontColor(1,1,1);

		//fontShadowColor(1,1,1);
		//fontShadowColor(0.5,0.5,0.5);
		//fontShadowColor(1,0.6,0.6);
		
		//fontGradientColor(0,0,0);
		//fontGradientColor(1,1,1);
		
		//fontShadow();
		//fontGradient();
	}
}

//////////////////////////////////////////////////////////////////////////////////////
//
// NOUVEAU THREAD - LE VOLUME DE LA MUSIQUE AUGMENTE
//
//
void FmodStreamFadingIn(void* dummy)
{
	int volume = 0;

	do
	{
		volume += 5;
		FSOUND_SetVolume(int_sound_mp3, volume);
		Sleep(30);
	}
	while(volume <= 255);

//	_endthread();
}


//////////////////////////////////////////////////////////////////////////////////////
//
// SON LORS DE L'APPUI SUR UNE Touch D'OPTION
//
//
void key_sound()
{
	int_sound = FSOUND_PlaySoundEx(FSOUND_FREE,sound_key, dspUnit, false);
	FSOUND_SetVolume(int_sound,200);
}

//////////////////////////////////////////////////////////////////////////////////////
//
// Check an OpenGL Extension. By Henry Goffin.
//
//
// WGLisExtensionSupported: This Is A Form Of The Extension For WGL
bool WGLisExtensionSupported(const char *extension)
{
	const size_t extlen = strlen(extension);
	const char *supported = NULL;

	// Try To Use wglGetExtensionStringARB On Current DC, If Possible
	PROC wglGetExtString = wglGetProcAddress("wglGetExtensionsStringARB");

	if (wglGetExtString)
		supported = ((char*(__stdcall*)(HDC))wglGetExtString)(wglGetCurrentDC());

	// If That Failed, Try Standard Opengl Extensions String
	if (supported == NULL)
		supported = (char*)glGetString(GL_EXTENSIONS);

	// If That Failed Too, Must Be No Extensions Supported
	if (supported == NULL)
		return false;

	// Begin Examination At Start Of String, Increment By 1 On False Match
	for (const char* p = supported; ; p++)
	{
		// Advance p Up To The Next Possible Match
		p = strstr(p, extension);

		if (p == NULL)
			return false;															// No Match

		// Make Sure That Match Is At The Start Of The String Or That
		// The Previous Char Is A Space, Or Else We Could Accidentally
		// Match "wglFunkywglExtension" With "wglExtension"

		// Also, Make Sure That The Following Character Is Space Or NULL
		// Or Else "wglExtensionTwo" Might Match "wglExtension"
		if ((p==supported || p[-1]==' ') && (p[extlen]=='\0' || p[extlen]==' '))
			return true;															// Match
	}
}

//////////////////////////////////////////////////////////////////////////////////////
//
// FSAA : Anti-aliasing (ARB Multi Sample) Author: Colt "MainRoach" McAnlis
//
//
// InitMultisample: Used To Query The Multisample Frequencies

#define WGL_SAMPLE_BUFFERS_ARB		 0x2041
#define WGL_SAMPLES_ARB			     0x2042

bool InitMultisample(HINSTANCE hInstance,HWND hWnd,PIXELFORMATDESCRIPTOR pfd)
{  
	if (!game_option_anti_aliasing) // l'option anti-aliasing n'a pas choisie
		return false;

	// See If The String Exists In WGL!
	if (!WGLisExtensionSupported("WGL_ARB_multisample"))
	{
		arbMultisampleSupported=false;
		return false;
	}

	// Get Our Pixel Format
	PFNWGLCHOOSEPIXELFORMATARBPROC wglChoosePixelFormatARB = (PFNWGLCHOOSEPIXELFORMATARBPROC)wglGetProcAddress("wglChoosePixelFormatARB");	
	if (!wglChoosePixelFormatARB) 
	{
		arbMultisampleSupported=false;
		return false;
	}

	// Get Our Current Device Context
	HDC hDC = GetDC(hWnd);

	int		pixelFormat;
	int		valid;
	UINT	numFormats;
	float	fAttributes[] = {0,0};

	// These Attributes Are The Bits We Want To Test For In Our Sample
	// Everything Is Pretty Standard, The Only One We Want To 
	// Really Focus On Is The SAMPLE BUFFERS ARB And WGL SAMPLES
	// These Two Are Going To Do The Main Testing For Whether Or Not
	// We Support Multisampling On This Hardware.
	int iAttributes[] =
	{
		WGL_DRAW_TO_WINDOW_ARB,GL_TRUE,
		WGL_SUPPORT_OPENGL_ARB,GL_TRUE,
		WGL_ACCELERATION_ARB,WGL_FULL_ACCELERATION_ARB,
		WGL_COLOR_BITS_ARB,24,	//32 work
		WGL_ALPHA_BITS_ARB,8,	//8
		WGL_DEPTH_BITS_ARB,16,	//8 work
		WGL_STENCIL_BITS_ARB,0,
		WGL_DOUBLE_BUFFER_ARB,GL_TRUE,
		WGL_SAMPLE_BUFFERS_ARB,GL_TRUE,
		WGL_SAMPLES_ARB,4,
		0,0
	};
	
	iAttributes[19] = 2; // 2 samples

	// First We Check To See If We Can Get A Pixel Format For 4 Samples
	valid = wglChoosePixelFormatARB(hDC,iAttributes,fAttributes,1,&pixelFormat,&numFormats);
 
	// If We Returned True, And Our Format Count Is Greater Than 1
	if (valid && numFormats >= 1)
	{
		arbMultisampleSupported = true;
		arbMultisampleFormat = pixelFormat;
		return arbMultisampleSupported;
	}

	// Our Pixel Format With 4 Samples Failed, Test For 2 Samples
	iAttributes[19] = 2;
	valid = wglChoosePixelFormatARB(hDC,iAttributes,fAttributes,1,&pixelFormat,&numFormats);
	if (valid && numFormats >= 1)
	{
		arbMultisampleSupported = true;
		arbMultisampleFormat = pixelFormat;	 
		return arbMultisampleSupported;
	}
	  
	// Return The Valid Format
	return  arbMultisampleSupported;
}

//////////////////////////////////////////////////////////////////////////////////////
//
// FONCTIONS => Affiche une erreur dans une MessageBox et quitte le programme
//
//
void DisplayMsgBoxErrorAndExit(LPCTSTR File, int Line, LPCTSTR Message)
{
	TCHAR ErrorBuf[512];

	// Create the string.
	wsprintf(ErrorBuf, "Sorry, an error has occurred...\n\n%s\nFile %s, line %d, version %s\n\nPlease check http://spaceinvadersgl.sourceforge.net to grab the last release.\nIf the error persists email mathieu at maya75@users.sourceforge.net.", 
		Message, File, Line, game_version);

	// Dump the error to the debugger, if present.
	OutputDebugString(ErrorBuf);

	// Display the error using MessageBox.
	if (Message != "")
		MessageBox(NULL, ErrorBuf, "Message", MB_OK | MB_TOPMOST);

	//exit(1);
}

//////////////////////////////////////////////////////////////////////////////////////
//
// FONCTIONS => Affiche le dernier message d'erreur sur la sortie de Debug
//
//
void DisplayLastErrorDebug(void)
{
	LPVOID lpMsgBuf;
	FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS, NULL, GetLastError(), MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), (LPTSTR) &lpMsgBuf, 0, NULL);
	OutputDebugString((LPCTSTR)lpMsgBuf);
	LocalFree(lpMsgBuf);
}